#****************************************************************************
#**
#**  File     :  /cdimage/lua/modules/shield.lua
#**
#**  Summary  :  Shield for units and structures.
#**
#**  Copyright  2005 Gas Powered Games, Inc.  All rights reserved.
#****************************************************************************
local Entity = import('entity.lua').Entity
local EffectTemplate = import('EffectTemplates.lua')
local Util = import('utilities.lua')

Shield = Class(moho.shield_methods,Entity) {

    ShieldVerticalOffset = -1,

    # Faction Mesh Blueprints
    ShieldMesh = {
        Aeon = {
            Mesh = '/effects/entities/AeonShield01/AeonShield01_mesh',
            MeshZ = '/effects/entities/Shield01/Shield01z_mesh',
            ImpactMesh = '/effects/entities/ShieldSection01/ShieldSection01_mesh',
            ImpactEffects = EffectTemplate.AeonShieldHit01,
        },
        Cybran = {
            Mesh = '/effects/entities/CybranShield01/CybranShield01_mesh',
            MeshZ = '/effects/entities/Shield01/Shield01z_mesh',
            ImpactMesh = '/effects/entities/ShieldSection01/ShieldSection01b_mesh',
            ImpactEffects = EffectTemplate.CybranShieldHit01,
        },
        UEF = {
            Mesh = '/effects/entities/Shield01/Shield01_mesh',
            MeshZ = '/effects/entities/Shield01/Shield01z_mesh',
            ImpactMesh = '/effects/entities/ShieldSection01/ShieldSection01_mesh',
            ImpactEffects = EffectTemplate.UEFShieldHit01,
        },
    },

    __init = function(self,spec)
        _c_CreateShield(self,spec)
    end,

    OnCreate = function(self,spec)
        self.Trash = TrashBag()
        self.Owner = spec.Owner
        self.FactionName = spec.FactionName
		self:SetSize(spec.Size)

        self:SetMaxHealth(spec.ShieldMaxHealth)
        self:SetHealth(self,spec.ShieldMaxHealth)

        # Show our 'lifebar'
        if not self.Owner:IsUnitState('Building') and
           not self.Owner:IsUnitState('Enhancing') and
           not self.Owner:IsUnitState('Repairing') then
            self.Owner:SetWorkProgress(1.0)
        end

        self:SetRechargeTime(spec.ShieldRechargeTime or 5, spec.ShieldEnergyDrainRechargeTime or 5)
        self:SetVerticalOffset(spec.ShieldVerticalOffset)

        self:SetVizToFocusPlayer('Always')
        self:SetVizToEnemies('Intel')
        self:SetVizToAllies('Always')
        self:SetVizToNeutrals('Intel')

        self:AttachBoneTo(-1,spec.Owner,-1)

        self:SetShieldRegenRate(spec.ShieldRegenRate)
        self:SetShieldRegenStartTime(spec.ShieldRegenStartTime)

		self.OffHealth = -1

        ChangeState(self, self.OnState)
    end,

    SetRechargeTime = function(self, rechargeTime, energyRechargeTime)
        self.ShieldRechargeTime = rechargeTime
        self.ShieldEnergyDrainRechargeTime = energyRechargeTime
    end,

    SetVerticalOffset = function(self, offset)
        self.ShieldVerticalOffset = offset
    end,

    SetSize = function(self, size)
        self.Size = size
    end,
    
    ModSize = function(self, sizemod)
        self.Size = self.Size + sizemod
        LOG("*Debug:Size=",self.Size)
    end,

    SetShieldRegenRate = function(self, rate)
        self.RegenRate = rate
    end,
	
    ModShieldRegenRate = function(self, ratemod)
        self.RegenRate = self.RegenRate + ratemod
        LOG("*Debug:RegenRate=",self.RegenRate)
    end,
    
    ModShieldHealth  = function(self, healthmod)
    	self:SetMaxHealth(self:GetMaxHealth() + healthmod) 
        LOG("*Debug:MaxHealth=",self:GetMaxHealth())
    end,
    
    SetShieldRegenStartTime = function(self, time)
        self.RegenStartTime = time
    end,

    # Note, this is called by native code to calculate spillover damage. The
    # damage logic will subtract this value from any damage it does to units
    # under the shield. The default is to always absorb as much as possible
    # but the reason this function exists is to allow flexible implementations
    # like shields that only absorb partial damage (like armor).
    OnGetDamageAbsorption = function(self,instigator,amount,type)
        #LOG('absorb: ', math.min( self:GetHealth(), amount ))
        return math.min( self:GetHealth(), amount )
    end,

    OnCollisionCheckWeapon = function(self, firingWeapon)
        local collide = firingWeapon:GetBlueprint().CollideFriendly
        if collide == false then
            if self:GetArmy() == firingWeapon.unit:GetArmy() then
                return false
            end
        end
        return true
    end,

    OnDamage =  function(self,instigator,amount,vector,type)
        local absorbed = self:OnGetDamageAbsorption(instigator,amount,type)
        self:AdjustHealth(instigator, -absorbed)

        if not self.Owner:IsUnitState('Building') and
           not self.Owner:IsUnitState('Enhancing') and
           not self.Owner:IsUnitState('Repairing') then
            self.Owner:SetWorkProgress(self:GetHealth() / self:GetMaxHealth())
        end

        #LOG('Shield Health: ' .. self:GetHealth())
        if self.RegenThread then
           KillThread(self.RegenThread)
           self.RegenThread = nil
        end
        if self:GetHealth() <= 0 then
            ChangeState(self, self.DamageRechargeState)
        else
            ForkThread(self.CreateImpactEffect, self, vector)
            if self.RegenRate > 0 then
                self.RegenThread = ForkThread(self.RegenStartThread, self)
                self.Owner.Trash:Add(self.RegenThread)
            end
        end
    end,

    RegenStartThread = function(self)
        WaitSeconds(self.RegenStartTime)
        while self:GetHealth() < self:GetMaxHealth() do
            self:AdjustHealth(self.Owner, self.RegenRate)

            if not self.Owner:IsUnitState('Building') and
               not self.Owner:IsUnitState('Enhancing') and
               not self.Owner:IsUnitState('Repairing') then
                self.Owner:SetWorkProgress(self:GetHealth() / self:GetMaxHealth())
            end

            WaitSeconds(1)
        end
    end,

    CreateImpactEffect = function(self, vector)
        local army = self:GetArmy()
		local OffsetLength = Util.GetVectorLength(vector)

		local ImpactMesh = Entity { Owner = self.Owner }
		Warp( ImpactMesh, self:GetPosition())
		ImpactMesh:SetMesh(self.ShieldMesh[self.FactionName].ImpactMesh)
		ImpactMesh:SetDrawScale(self.Size)
		ImpactMesh:SetOrientation(OrientFromDir(Vector(-vector.x,-vector.y,-vector.z)),true)

		for k, v in self.ShieldMesh[self.FactionName].ImpactEffects do
			CreateEmitterAtBone( ImpactMesh, -1, army, v ):OffsetEmitter(0,0,OffsetLength)
		end

		WaitSeconds(5)

		ImpactMesh:Destroy()
    end,

    OnDestroy = function(self)
		self:SetMesh('')
		if self.MeshZ != nil then
			self.MeshZ:Destroy()
			self.MeshZ = nil
		end
        ChangeState(self, self.DeadState)
    end,

    # Return true to process this collision, false to ignore it.
    OnCollisionCheck = function(self,other)
        if other:GetArmy() == -1 then
            return false
        end

        return IsEnemy(self:GetArmy(),other:GetArmy())
    end,

    TurnOn = function(self)
        ChangeState(self, self.OnState)
    end,

    TurnOff = function(self)
        ChangeState(self, self.OffState)
    end,

    IsOn = function(self)
        return false
    end,

    RemoveShield = function(self)
        self:SetCollisionShape('None')

		self:SetMesh('')
		if self.MeshZ != nil then
			self.MeshZ:Destroy()
			self.MeshZ = nil
		end

		self.Owner:OnShieldDisabled()
    end,

    CreateShieldMesh = function(self)
		self:SetCollisionShape( 'Sphere', 0, 0, 0, self.Size/2)

		self:SetMesh(self.ShieldMesh[self.FactionName].Mesh)
		self:SetParentOffset(Vector(0,self.ShieldVerticalOffset,0))
		self:SetDrawScale(self.Size)

		if self.MeshZ == nil then
			self.MeshZ = Entity { Owner = self.Owner }
			self.MeshZ:SetMesh(self.ShieldMesh[self.FactionName].MeshZ)
            Warp( self.MeshZ, self.Owner:GetPosition() )
			self.MeshZ:SetDrawScale(self.Size)
			self.MeshZ:AttachBoneTo(-1,self.Owner,-1)
			self.MeshZ:SetParentOffset(Vector(0,self.ShieldVerticalOffset,0))

			self.MeshZ:SetVizToFocusPlayer('Always')
			self.MeshZ:SetVizToEnemies('Intel')
			self.MeshZ:SetVizToAllies('Always')
			self.MeshZ:SetVizToNeutrals('Intel')
		end
    end,

    OnState = State {
        Main = function(self)

            # Set initial health of the shield
            if self.OffHealth >= 0 then
                self:SetHealth(self, self.OffHealth)
                self.OffHealth = -1
			else
                self:SetHealth(self, self:GetMaxHealth())
            end

            if not self.Owner:IsUnitState('Building') and
               not self.Owner:IsUnitState('Enhancing') and
               not self.Owner:IsUnitState('Repairing') then
                self.Owner:SetWorkProgress(self:GetHealth() / self:GetMaxHealth())
            end

            self.Owner:OnShieldEnabled()
			self:CreateShieldMesh()

            WaitSeconds(0.5)
            local fraction = self.Owner:GetResourceConsumed()
            local on = true
            local test = false
            while on do
                WaitTicks(1)

                if not self.Owner:IsUnitState('Building') and
                   not self.Owner:IsUnitState('Enhancing') and
                   not self.Owner:IsUnitState('Repairing') then
                    self.Owner:SetWorkProgress(self:GetHealth() / self:GetMaxHealth())
                end
                
                fraction = self.Owner:GetResourceConsumed()
                if fraction != 1 then
                    if test then
                        on = false
                    else
                        test = true
                    end
                else
                    on = true
                    test = false
                end
            end
            ChangeState(self, self.EnergyDrainRechargeState)
        end,

        IsOn = function(self)
            return true
        end,

    },

    OffState = State {
        Main = function(self)

            if self.RegenThread then
                KillThread(self.RegenThread)
                self.RegenThread = nil
            end

      		self.OffHealth = self:GetHealth()

            if not self.Owner:IsUnitState('Building') and
               not self.Owner:IsUnitState('Enhancing') and
               not self.Owner:IsUnitState('Repairing') then
                self.Owner:SetWorkProgress(0)
            end

            self:RemoveShield()

            WaitSeconds(1)

            while self.OffHealth < self:GetMaxHealth() do
                self.OffHealth = math.min( self:GetMaxHealth(), self.OffHealth + ( self:GetMaxHealth() / self.ShieldRechargeTime ) )
                WaitSeconds(1)
            end
        end,
    },

    DamageRechargeState = State {
        Main = function(self)
            self:RemoveShield()
            WaitSeconds(self.ShieldRechargeTime)
            ChangeState(self, self.OnState)
        end
    },

    EnergyDrainRechargeState = State {
        Main = function(self)
            self:RemoveShield()
            WaitSeconds(self.ShieldEnergyDrainRechargeTime)
            if not self.Owner:IsUnitState('Attached') then
                ChangeState(self, self.OnState)
            else
                ChangeState(self, self.OffState)
            end
        end
    },

    DeadState = State {
        Main = function(self)
        end,
    },
}

UnitShield = Class(Shield){

    OnCreate = function(self,spec)
        self.Trash = TrashBag()
        self.Owner = spec.Owner
        self.FactionName = spec.FactionName
        self.CollisionSizeX = spec.CollisionSizeX or 1
		self.CollisionSizeY = spec.CollisionSizeY or 1
		self.CollisionSizeZ = spec.CollisionSizeZ or 1
		self.CollisionCenterX = spec.CollisionCenterX or 0
		self.CollisionCenterY = spec.CollisionCenterY or 0
		self.CollisionCenterZ = spec.CollisionCenterZ or 0
		self.OwnerShieldMesh = spec.OwnerShieldMesh or ''

		self:SetSize(spec.Size)

        self:SetMaxHealth(spec.ShieldMaxHealth)
        self:SetHealth(self,spec.ShieldMaxHealth)

        # Show our 'lifebar'
        if not self.Owner:IsUnitState('Building') and
           not self.Owner:IsUnitState('Enhancing') and
           not self.Owner:IsUnitState('Repairing') then
            self.Owner:SetWorkProgress(1.0)
        end

        self:SetRechargeTime(spec.ShieldRechargeTime or 5, spec.ShieldEnergyDrainRechargeTime or 5)
        self:SetVerticalOffset(spec.ShieldVerticalOffset)

        self:SetVizToFocusPlayer('Always')
        self:SetVizToEnemies('Intel')
        self:SetVizToAllies('Always')
        self:SetVizToNeutrals('Always')

        self:AttachBoneTo(-1,spec.Owner,-1)

        self:SetShieldRegenRate(spec.ShieldRegenRate)
        self:SetShieldRegenStartTime(spec.ShieldRegenStartTime)

        ChangeState(self, self.OnState)
    end,

    CreateImpactEffect = function(self, vector)
        local army = self:GetArmy()
		local OffsetLength = Util.GetVectorLength(vector)
		local ImpactEnt = Entity { Owner = self.Owner }

		Warp( ImpactEnt, self:GetPosition())
		ImpactEnt:SetOrientation(OrientFromDir(Vector(-vector.x,-vector.y,-vector.z)),true)

		for k, v in self.ShieldMesh[self.FactionName].ImpactEffects do
			CreateEmitterAtBone( ImpactEnt, -1, army, v ):OffsetEmitter(0,0,OffsetLength)
		end
		WaitSeconds(1)

		ImpactEnt:Destroy()
    end,

    CreateShieldMesh = function(self)
		self:SetCollisionShape( 'Box', self.CollisionCenterX, self.CollisionCenterY, self.CollisionCenterZ, self.CollisionSizeX, self.CollisionSizeY, self.CollisionSizeZ)
		self.Owner:SetMesh(self.OwnerShieldMesh,true)
    end,

    RemoveShield = function(self)
        self:SetCollisionShape('None')
		self.Owner:SetMesh(self.Owner:GetBlueprint().Display.MeshBlueprint, true)
		self.Owner:OnShieldDisabled()
    end,

    OnDestroy = function(self)
		self.Owner:SetMesh(self.Owner:GetBlueprint().Display.MeshBlueprint, true)
        ChangeState(self, self.DeadState)
    end,

}