--*****************************************************************************
--* File: lua/modules/ui/game/construction.lua
--* Author: Chris Blackwell
--* Summary: Construction management UI
--*
--* Copyright  2005 Gas Powered Games, Inc.  All rights reserved.
--*****************************************************************************

local UIUtil = import('/lua/modules/ui/uiutil.lua')
local LayoutHelpers = import('/lua/modules/maui/layouthelpers.lua')
local Group = import('/lua/modules/maui/group.lua').Group
local Bitmap = import('/lua/modules/maui/bitmap.lua').Bitmap
local Grid = import('/lua/modules/maui/grid.lua').Grid
local Checkbox = import('/lua/modules/maui/checkbox.lua').Checkbox
local Button = import('/lua/modules/maui/button.lua').Button
local GameCommon = import('/lua/modules/ui/game/gamecommon.lua')
local GameMain = import('/lua/modules/ui/game/gamemain.lua')
local RadioGroup = import('/lua/modules/maui/mauiutil.lua').RadioGroup
local Queue = import('/lua/modules/ui/game/queue.lua')
local Tooltip = import('/lua/modules/ui/game/tooltip.lua')

local unitGridPages = {
    RULEUTL_Basic = {Order = 0, Label = "<LOC CONSTRUCT_0000>T1"},
    RULEUTL_Advanced = {Order = 1, Label = "<LOC CONSTRUCT_0001>T2"},
    RULEUTL_Secret = {Order = 2, Label = "<LOC CONSTRUCT_0002>T3"},
    RULEUTL_Experimental = {Order = 3, Label = "<LOC CONSTRUCT_0003>Exp"},
    RULEUTL_Munition = {Order = 4, Label = "<LOC CONSTRUCT_0004>Munition"},   -- note that this doesn't exist yet
}

-- these are external controls used for positioning, so don't add them to our local control table
controlClusterGroup = false
mfdControl = false
ordersControl = false
queueControl = false
local filteredFactories = {}
local filteredTech1Units = {}
local filteredTech2Units = {}
local filteredTech3Units = {}
local filteredTech4Units = {}
local filteredEnhancements = {}
local filteredTechTable = {}
local techLevelSelected = filteredTechTable[1]
local tempBuildableCategories = false
local currentSelection = false
local techEnhTab = false
local buildingEnhancements = false
local currentEntityList = false
local mouseoverDisplay = false
local enhancementBool = false
local layoutVar = false

local newTechUnits = {}

controls = {
    unitGrid = false,
    constructionGroup = false,
    scrollUnitsFwdButton = false,
    scrollUnitsBackButton = false,
    unitsBGCapMin = false,
    unitsBGMiddle = false,
    unitsBGCapMax = false,
    techRadioGroup = false,
    scrollBG = false,
    enhanceButton = false,
    tech1Tab = false,
    tech2Tab = false,
    tech3Tab = false,
    tech4Tab = false,
    tech5Tab = false,
    techTabControlTable = {},
    tech2disabledOverGroup = false,
    tech3disabledOverGroup = false,
    tech4disabledOverGroup = false,
}


local whatIfBuilder = nil
local whatIfBlueprintID = nil
local selectedwhatIfBuilder = nil
local selectedwhatIfBlueprintID = nil

function GetTableSize(table)
    index = 0
    if table then
        for i, v in table do
            index = index + 1
        end
    end
    return index
end

function ClearWhatIfModifier()
    whatIfBuilder = nil
    whatIfBlueprintID = nil
    selectedwhatIfBuilder = nil
    selectedwhatIfBlueprintID = nil
end

function GetCurrentWhatIfModifier()
    local consumeEnergy = 0
    local consumeMass = 0
    local storeEnergy = 0
    local storeMass = 0

    if whatIfBuilder != nil and whatIfBlueprintID != nil then
        local targetBP = __blueprints[whatIfBlueprintID]
        local time, energy, mass = import('/lua/modules/game.lua').GetConstructEconomyModel(whatIfBuilder, targetBP.Economy)
        if time > 0 then
            consumeEnergy = -energy / time
            consumeMass = -mass / time
        end
        storeEnergy = targetBP.Economy.StorageEnergy or 0
        storeMass = targetBP.Economy.StorageMass or 0
    elseif selectedwhatIfBuilder != nil and selectedwhatIfBlueprintID != nil then
        local targetBP = __blueprints[selectedwhatIfBlueprintID]
        local time, energy, mass = import('/lua/modules/game.lua').GetConstructEconomyModel(selectedwhatIfBuilder, targetBP.Economy)
        if time > 0 then
            consumeEnergy = -energy / time
            consumeMass = -mass / time
        end
        storeEnergy = targetBP.Economy.StorageEnergy or 0
        storeMass = targetBP.Economy.StorageMass or 0
    end

    return { consumeEnergy,
             consumeMass,
             storeEnergy,
             storeMass }
end

function SetConstructionState(state)
    controls.constructionGroup.Mini(state)
    controls.unitGrid:DeleteAndDestroyAll(true)
    ClearSessionExtraSelectList()
    if state then
        currentSelection = {}
    end
end

function CreateElements()
    if not controls.constructionGroup then
        controls.constructionGroup = Group(controlClusterGroup, "controlClusterGroup")
    end
    if not controls.unitsBGCapMin then
        controls.unitsBGCapMin = Bitmap(controls.constructionGroup)
    end
    if not controls.unitsBGCapMax then
        controls.unitsBGCapMax = Bitmap(controls.constructionGroup)
    end
    if not controls.unitsBGMiddle then
        controls.unitsBGMiddle = Bitmap(controls.constructionGroup)
    end
    
    if not controls.tech1Tab then
        controls.tech1Tab = Group(controls.constructionGroup, "tech1TabGroup")
        controls.tech1Tab.Top:Set(0)
        controls.tech1Tab.Left:Set(0)
        controls.tech1Tab.Right:Set(1)
        controls.tech1Tab.Bottom:Set(1)
    end
    
    if not controls.tech2Tab then
        controls.tech2Tab = Group(controls.constructionGroup, "tech2TabGroup")
        controls.tech2Tab.Top:Set(0)
        controls.tech2Tab.Left:Set(0)
        controls.tech2Tab.Right:Set(1)
        controls.tech2Tab.Bottom:Set(1)
    end
    
    if not controls.tech3Tab then
        controls.tech3Tab = Group(controls.constructionGroup, "tech3TabGroup")
        controls.tech3Tab.Top:Set(0)
        controls.tech3Tab.Left:Set(0)
        controls.tech3Tab.Right:Set(1)
        controls.tech3Tab.Bottom:Set(1)
    end
    
    if not controls.tech4Tab then
        controls.tech4Tab = Group(controls.constructionGroup, "tech4TabGroup")
        controls.tech4Tab.Top:Set(0)
        controls.tech4Tab.Left:Set(0)
        controls.tech4Tab.Right:Set(1)
        controls.tech4Tab.Bottom:Set(1)
    end
    
    if not controls.tech1Tab.checkbox then
        controls.tech1Tab.checkbox = UIUtil.CreateCheckbox(controls.tech1Tab, nil, nil, nil, nil, nil, nil, "UI_Tech_Level_Click", "UI_Tech_Level_Rollover")
        controls.tech1Tab.icon = Bitmap(controls.tech1Tab, UIUtil.SkinnableFile('/game/tech-tabs_btn/tech-level-1_icon_bmp.dds'))
    end
    
    if not controls.tech2Tab.checkbox then
        controls.tech2Tab.checkbox = UIUtil.CreateCheckbox(controls.tech2Tab, nil, nil, nil, nil, nil, nil, "UI_Tech_Level_Click", "UI_Tech_Level_Rollover")
        controls.tech2Tab.icon = Bitmap(controls.tech2Tab, UIUtil.SkinnableFile('/game/tech-tabs_btn/tech-level-2_icon_bmp.dds'))
    end
    
    if not controls.tech3Tab.checkbox then
        controls.tech3Tab.checkbox = UIUtil.CreateCheckbox(controls.tech3Tab, nil, nil, nil, nil, nil, nil, "UI_Tech_Level_Click", "UI_Tech_Level_Rollover")
        controls.tech3Tab.icon = Bitmap(controls.tech3Tab, UIUtil.SkinnableFile('/game/tech-tabs_btn/tech-level-3_icon_bmp.dds'))
    end
    
    if not controls.tech4Tab.checkbox then
        controls.tech4Tab.checkbox = UIUtil.CreateCheckbox(controls.tech4Tab, nil, nil, nil, nil, nil, nil, "UI_Tech_Level_Click", "UI_Tech_Level_Rollover")
        controls.tech4Tab.icon = Bitmap(controls.tech4Tab, UIUtil.SkinnableFile('/game/tech-tabs_btn/experimental_icon_bmp.dds'))
    end
    
    if not controls.unitGrid then
        controls.unitGrid = Grid(controls.unitsBGMiddle, 62, 62)
        controls.unitGrid:SetName("Unit Grid")
    end
    
    if not controls.tech2disabledOverGroup then
        controls.tech2disabledOverGroup = Group(controls.constructionGroup)
    end
    
    if not controls.tech3disabledOverGroup then
        controls.tech3disabledOverGroup = Group(controls.constructionGroup)
    end
    
    if not controls.tech4disabledOverGroup then
        controls.tech4disabledOverGroup = Group(controls.constructionGroup)
    end
    
    if not controls.tech1Tab.label then
        controls.tech1Tab.label = UIUtil.CreateText(controls.tech1Tab, "", 16, UIUtil.bodyFont)
        controls.tech1Tab.label.Top:Set(0)
        controls.tech1Tab.label.Left:Set(0)
        controls.tech1Tab.label.Right:Set(1)
        controls.tech1Tab.label.Bottom:Set(1)
    end
    
    if not controls.tech2Tab.label then
        controls.tech2Tab.label = UIUtil.CreateText(controls.tech2Tab, "", 16, UIUtil.bodyFont)
        controls.tech2Tab.label.Top:Set(0)
        controls.tech2Tab.label.Left:Set(0)
        controls.tech2Tab.label.Right:Set(1)
        controls.tech2Tab.label.Bottom:Set(1)
    end
    
    if not controls.tech3Tab.label then
        controls.tech3Tab.label = UIUtil.CreateText(controls.tech3Tab, "", 16, UIUtil.bodyFont)
        controls.tech3Tab.label.Top:Set(0)
        controls.tech3Tab.label.Left:Set(0)
        controls.tech3Tab.label.Right:Set(1)
        controls.tech3Tab.label.Bottom:Set(1)
    end
    
    if not controls.tech4Tab.label then
        controls.tech4Tab.label = UIUtil.CreateText(controls.tech4Tab, "", 16, UIUtil.bodyFont)
        controls.tech4Tab.label.Top:Set(0)
        controls.tech4Tab.label.Left:Set(0)
        controls.tech4Tab.label.Right:Set(1)
        controls.tech4Tab.label.Bottom:Set(1)
    end
    
    controls.techTabControlTable[1] = controls.tech1Tab
    controls.techTabControlTable[2] = controls.tech2Tab
    controls.techTabControlTable[3] = controls.tech3Tab
    controls.techTabControlTable[4] = controls.tech4Tab
    
    if not controls.scrollUnitsFwdButton then
        controls.scrollUnitsFwdButton = UIUtil.CreateButton(controls.unitsBGCapMin
            , '/game/lg-vert_scr/arrow-up_scr_up.dds'
            , '/game/lg-vert_scr/arrow-up_scr_down.dds'
            , '/game/lg-vert_scr/arrow-up_scr_over.dds'
            , '/game/lg-vert_scr/arrow-up_scr_dis.dds'
            , nil, nil, nil, nil, 'UI_Arrow_Click')
    end
    
    if not controls.scrollUnitsBackButton then
        controls.scrollUnitsBackButton = UIUtil.CreateButton(controls.unitsBGCapMin
            , '/game/lg-vert_scr/arrow-up_scr_up.dds'
            , '/game/lg-vert_scr/arrow-up_scr_down.dds'
            , '/game/lg-vert_scr/arrow-up_scr_over.dds'
            , '/game/lg-vert_scr/arrow-up_scr_dis.dds'
            , nil, nil, nil, nil, 'UI_Arrow_Click')
    end
    
    if not controls.enhanceButton then
        controls.enhanceButton = Button(controls.unitsBGCapMin, nil, nil, nil, nil, "UI_Enhancements_Click", "UI_Enhancements_Rollover")
    end
end

function CommonLogic()
    controls.techRadioGroup = RadioGroup(controls.techTabControlTable, 1)
    controls.scrollUnitsBackButton:UseAlphaHitTest(false)
    controls.scrollUnitsFwdButton:UseAlphaHitTest(false)

    if controls.unitGrid then
        controls.unitGrid.Width.OnDirty = function(var) PopulateGridButtons() end
        controls.unitGrid.Height.OnDirty = function(var) PopulateGridButtons() end
    end
    controls.tech2disabledOverGroup.HandleEvent = function(self, event)
        if event.Type == 'MouseEnter' then
            Tooltip.CreateMouseoverDisplay(self, "tech_level_2_cons_tab_dis", nil, true)
        elseif event.Type == 'MouseExit' then
            Tooltip.DestroyMouseoverDisplay()
        end
    end 
    controls.tech3disabledOverGroup.HandleEvent = function(self, event)
        if event.Type == 'MouseEnter' then
            Tooltip.CreateMouseoverDisplay(self, "tech_level_3_cons_tab_dis", nil, true)
        elseif event.Type == 'MouseExit' then
            Tooltip.DestroyMouseoverDisplay()
        end
    end 
    controls.tech4disabledOverGroup.HandleEvent = function(self, event)
        if event.Type == 'MouseEnter' then
            Tooltip.CreateMouseoverDisplay(self, "tech_level_4_cons_tab_dis", nil, true)
        elseif event.Type == 'MouseExit' then
            Tooltip.DestroyMouseoverDisplay()
        end
    end 
    controls.techRadioGroup.OnChoose = function(self, index, item)
        techLevelSelected = filteredTechTable[index]
        buildingEnhancements = false
        local sound = Sound({Bank = 'Interface', Cue = 'UI_Tech_Level_Click'})
        PlaySound(sound)
        PopulateGridButtons()
    end
    controls.techRadioGroup:SetCheckboxEventHandler(function(self, event)
        if event.Type == 'MouseEnter' then
            Tooltip.CreateMouseoverDisplay(self, "tech_level_"..self._radioGroupIndex.."_cons_tab")
        elseif event.Type == 'MouseExit' then
            Tooltip.DestroyMouseoverDisplay()
        end
    end)
    controls.scrollUnitsBackButton.OnClick = function(self)
        controls.unitGrid:ScrollLines("Vert", -1)
    end
    controls.scrollUnitsFwdButton.OnClick = function(self)
        controls.unitGrid:ScrollLines("Vert", 1)
    end
    controls.enhanceButton.OnClick = function(self, modifiers)
        import('/lua/modules/ui/game/enhancements.lua').ToggleDialog(filteredEnhancements, currentSelection[1])
    end
    controls.enhanceButton.HandleEvent = function(self, event)
        if event.Type == 'MouseEnter' then
            Tooltip.CreateMouseoverDisplay(self, "enhance")
        elseif event.Type == 'MouseExit' then
            Tooltip.DestroyMouseoverDisplay()
        end
        Button.HandleEvent(self, event)
    end
end

function SetLayout(layout)
    if controls.unitGrid then
        controls.unitGrid.Width.OnDirty = nil
        controls.unitGrid.Height.OnDirty = nil
    end

    layoutVar = layout
    import(UIUtil.GetLayoutFilename('construction')).SetLayout()
    CommonLogic()
end

function SetupConstructionControl(parent, inMFDControl, inOrdersControl, inQueueControl)
    mfdControl = inMFDControl
    ordersControl = inOrdersControl
    controlClusterGroup = parent
    queueControl = inQueueControl
    
    CreateElements()

    SetLayout(UIUtil.currentLayout)

    return controls.constructionGroup
end

function CreateEnhancementButton(parent, enhancementName, enhancement)
    local tempBmpname = ""
    local tempEntityID = currentSelection[1]:GetBlueprint()

    if enhancement.Slot == "Back" then
        tempBmpName = tempEntityID.BlueprintId .. "-b-" .. enhancement.Icon
    elseif enhancement.Slot == "LCH" then
        tempBmpName = tempEntityID.BlueprintId .. "-la-" .. enhancement.Icon
    elseif enhancement.Slot == "RCH" then
        tempBmpName = tempEntityID.BlueprintId .. "-ra-" .. enhancement.Icon
    elseif enhancement.Slot == "Shield" then
        tempBmpName = tempEntityID.BlueprintId .. "-" .. enhancement.Icon
    else
    	tempBmpName = tempEntityID.BlueprintId .. "-".. enhancement.Slot .."-" .. enhancement.Icon
    end

    if string.sub(tempEntityID.BlueprintId, 2, 2) == 'a' then
        tempBmpName = '/dialogs/enhance-icons-aeon-' .. tempEntityID.BlueprintId .. '/'.. tempBmpName
    elseif string.sub(tempEntityID.BlueprintId, 2, 2) == 'e' then
        tempBmpName = '/dialogs/enhance-icons-uef-' .. tempEntityID.BlueprintId .. '/' .. tempBmpName
    elseif string.sub(tempEntityID.BlueprintId, 2, 2) == 'r' then
        tempBmpName = '/dialogs/enhance-icons-cybran-' .. tempEntityID.BlueprintId .. '/' .. tempBmpName
    end

    local bg = Bitmap(parent)   --todo put land/air/sea background here
    bg:SetSolidColor('black')
    bg.Width:Set(GameCommon.iconBmpWidth)
    bg.Height:Set(GameCommon.iconBmpHeight)
    local button = Button(bg,
        UIUtil.UIFile(tempBmpName .. '_btn_up.dds'),
        UIUtil.UIFile(tempBmpName .. '_btn_over.dds'),
        UIUtil.UIFile(tempBmpName .. '_btn_down.dds'),
        UIUtil.UIFile(tempBmpName .. '_btn_sel.dds'),
        "UI_Enhancements_Click", "UI_Enhancements_Rollover")
    --LayoutHelpers.FillParent(button, bg)
    button.Top:Set(bg.Top)
    button.Left:Set(function() return bg.Left() + 1 end)
    if layoutVar == 'minimal' then
        button.Height:Set(48)
        button.Width:Set(48)
    end

    button.OnClick = function(self, modifiers)
        IssueCommand("UNITCOMMAND_Enhance", enhancementName, true)
        import('/lua/modules/ui/game/enhancements.lua').ToggleDialog(filteredEnhancements, currentSelection[1], true)
    end

    button.HandleEvent = function(self, event)
        if event.Type == 'MouseEnter' then
            import('/lua/modules/ui/game/unitviewDetail.lua').ShowEnhancement(enhancement, tempEntityID.BlueprintId, enhancement.Icon, currentSelection[1])
        end
        if event.Type == 'MouseExit' then
            import('/lua/modules/ui/game/unitviewDetail.lua').Hide()
        end
        Button.HandleEvent(self, event)
    end

    return bg
end

local UPPER_GLOW_THRESHHOLD = .5
local LOWER_GLOW_THRESHHOLD = .1
local GLOW_SPEED = 2

-- not standard so use our own button create code
function CreateBuildableButton(parent, blueprintId, blueprint, unitList)
    local iconName, upIconName, downIconName, overIconName = GameCommon.GetCachedUnitIconFileNames(blueprint)
    local bg = Group(parent)
    bg.Width:Set(GameCommon.iconBmpWidth)
    bg.Height:Set(GameCommon.iconBmpHeight)
    local button = Button(bg, upIconName, downIconName, overIconName, upIconName, 'UI_Opt_Mini_Button_Click', 'UI_Opt_Mini_Button_Over')
    local selectedUnits = unitList
    button.Top:Set(bg.Top)
    button.Left:Set(function() return bg.Left() + 1 end)
    if layoutVar == 'minimal' then
        button.Height:Set(48)
        button.Width:Set(48)
    end
    button._blueprintId = blueprintId

    if newTechUnits and table.find(newTechUnits, blueprintId) then
        table.remove(newTechUnits, table.find(newTechUnits, blueprintId))
        button.NewInd = Bitmap(button, UIUtil.UIFile('/game/selection/selection_brackets_player_highlighted.dds'))
        button.NewInd.Height:Set(80)
        button.NewInd.Width:Set(80)
        LayoutHelpers.AtCenterIn(button.NewInd, button)
        button.NewInd:DisableHitTest()
        button.NewInd.Incrementing = false
        button.NewInd:SetNeedsFrameUpdate(true)
        button.NewInd.OnFrame = function(self, delta)
            local newAlpha = self:GetAlpha() - delta / 5
            if newAlpha < 0 then
                self:SetAlpha(0)
                self:SetNeedsFrameUpdate(false)
                return
            else
                self:SetAlpha(newAlpha)
            end
            if self.Incrementing then
                local newheight = self.Height() + delta * 100
                if newheight > 80 then
                    self.Height:Set(80)
                    self.Width:Set(80)
                    self.Incrementing = false
                else
                    self.Height:Set(newheight)
                    self.Width:Set(newheight)
                end
            else
                local newheight = self.Height() - delta * 100
                if newheight < 50 then
                    self.Height:Set(50)
                    self.Width:Set(50)
                    self.Incrementing = true
                else
                    self.Height:Set(newheight)
                    self.Width:Set(newheight)
                end
            end
        end
    end
    
    
    button.Glow = Bitmap(button)
    button.Glow:SetTexture(UIUtil.UIFile('/game/units_bmp/glow.dds'))
    --button.Glow:SetSolidColor('FF00FF00')
    --button.Glow.Height:Set(64)
    --button.Glow.Width:Set(64)
    LayoutHelpers.AtCenterIn(button.Glow, button)
    button.Glow:SetAlpha(0)
    button.Incrementing = true
    button.Glow.OnFrame = function(self, elapsedTime)
        local curAlpha = self:GetAlpha()
        if self.Incrementing then
            curAlpha = curAlpha + (elapsedTime * GLOW_SPEED)
            if curAlpha > UPPER_GLOW_THRESHHOLD then
                curAlpha = UPPER_GLOW_THRESHHOLD
                self.Incrementing = false
            end
        else
            curAlpha = curAlpha - (elapsedTime * GLOW_SPEED)
            if curAlpha < LOWER_GLOW_THRESHHOLD then
                curAlpha = LOWER_GLOW_THRESHHOLD
                self.Incrementing = true
            end
        end
        self:SetAlpha(curAlpha)
    end

    local unitBuildKeys = import('/lua/modules/ui/game/buildmode.lua').GetUnitKeys(selectedUnits[1]:GetBlueprint().BlueprintId, GetCurrentTechTab())
    if unitBuildKeys and unitBuildKeys[blueprint.BlueprintId] then
        button.KeyBG = Bitmap(button)
        button.KeyBG:SetSolidColor('AA000000')
        button.KeyBG.Height:Set(function() return button.Height() / 2 end)
        button.KeyBG.Width:Set(function() return button.Width() / 2 end)
        button.KeyBG.Right:Set(button.Right)
        button.KeyBG.Bottom:Set(button.Bottom)
        button.KeyBG.Depth:Set(function() return button.Glow.Depth() + 1 end)
        
        button.Key = UIUtil.CreateText(button.KeyBG, unitBuildKeys[blueprint.BlueprintId], 18, "Arial")
        LayoutHelpers.AtCenterIn(button.Key, button.KeyBG)
        button.Key:SetColor('ffff9900')
        
        button.KeyBG:DisableHitTest(true)
        button.KeyBG:Hide()
        button.KeyBG.OnHide = function(self)
            if import('/lua/modules/ui/game/buildmode.lua').IsInBuildMode() then
                return false
            else
                return true
            end
        end
    end
    
    button.OnClick = function(self, modifiers)
        local blueprint = __blueprints[self._blueprintId]
        local count = 1
        local performUpgrade = false

        selectedwhatIfBuilder = whatIfBuilder
        selectedwhatIfBlueprintID = whatIfBlueprintID

        if modifiers.Left then
            if modifiers.Ctrl == true and modifiers.Shift == nil then
                count = 5
            elseif modifiers.Shift == true and modifiers.Ctrl == nil then
                count = 5
            elseif modifiers.Shift == true and modifiers.Ctrl == true then
                count = 5
            end

            -- see if we are issuing an upgrade order
            if blueprint.General.UpgradesFrom == 'none' then
                performUpgrade = false
            else
                for i,v in selectedUnits do
                    if v then   -- it's possible that your unit will have died by the time this gets to it
                        local unitBp = v:GetBlueprint()
                        if blueprint.General.UpgradesFrom == unitBp.BlueprintId then
                            performUpgrade = true
                        elseif blueprint.General.UpgradesFrom == unitBp.General.UpgradesTo then
                            performUpgrade = true
                        end
                    end
                end
            end

            if performUpgrade then
                -- if the item to build is an upgrade.. then issue the upgrade command
                local clear = false
                
                -- I don't think this is needed anymore, since commanders use the enhancement system, not upgrades. 
                -- Please do not turn this back on without talking to dteh.
                --if blueprint.Physics.MotionType == 'RULEUMT_None' or modifiers.Shift then
                --    clear = false
                --end
                
                IssueBlueprintCommand("UNITCOMMAND_Upgrade", self._blueprintId, 1, clear)
            else
                if blueprint.Physics.MotionType == 'RULEUMT_None' or EntityCategoryContains(categories.NEEDMOBILEBUILD, self._blueprintId) then
                    -- stationary means it needs to be placed, so go in to build mobile mode
                    import('/lua/modules/ui/game/commandmode.lua').StartCommandMode("build", self._blueprintId)
                else
                    -- if the item to build can move, it must be built by a factory
                    --TODO -what about mobile factories?
                    IssueBlueprintCommand("UNITCOMMAND_BuildFactory", self._blueprintId, count)
                end
            end
        else
            import('/lua/modules/ui/game/queue.lua').RemovedLastQueuedUnit(self._blueprintId)
        end
    end

    button.HandleEvent = function(self, event)
        if event.Type == 'MouseEnter' then
            self.Glow:SetNeedsFrameUpdate(true)
            self.Glow:SetAlpha(LOWER_GLOW_THRESHHOLD)
            local blueprint = __blueprints[self._blueprintId]
            local builder = unitList[1]
            whatIfBuilder = builder
            whatIfBlueprintID = self._blueprintId
            Tooltip.CreateMouseoverDisplay(self, blueprint.Description)
            import('/lua/modules/ui/game/unitviewDetail.lua').Show(blueprint, builder, self._blueprintId)
        end
        if event.Type == 'MouseExit' then
            self.Glow:SetNeedsFrameUpdate(false)
            self.Glow:SetAlpha(0)
            whatIfBuilder = nil
            whatIfBlueprintID = nil
            Tooltip.DestroyMouseoverDisplay()
            import('/lua/modules/ui/game/unitviewDetail.lua').Hide()
        end
        Button.HandleEvent(self, event)
    end
    bg.button = button
    return bg
end

function ShowBuildModeKeys(show)
    local cols, rows = controls.unitGrid:GetDimensions()
    local items = cols * rows
    for i = 1, items do
        local curCol = math.mod(i - 1, cols) + 1 -- note the sub/add due to one based arrays
        local curRow = math.ceil(i / cols)
        if controls.unitGrid:GetItem(curCol, curRow) then
            local control = controls.unitGrid:GetItem(curCol, curRow)
            if control.button.KeyBG then
                if not control:IsHidden() then
                    control.button.KeyBG:SetHidden(not show)
                end
            end
        end
    end
end

-- not standard so use our own button create code
function CreateAttachedEntityButton(parent, blueprintId, blueprint, userUnit)
    local iconName, upIconName, downIconName, overIconName = GameCommon.GetCachedUnitIconFileNames(blueprint)
    local selectedIcon = '/textures/ui/common/icons/units/default_vehicle_icon.dds' -- Tempory icons
    local disabledIcon = '/textures/ui/common/icons/units/default_icon.dds'         -- Tempory icons
    local bg = Bitmap(parent)   --todo put land/air/sea background here
    local disabledButton = userUnit:HasUnloadCommandQueuedUp()
    bg:SetSolidColor('black')
    bg.Width:Set(GameCommon.iconBmpWidth)
    bg.Height:Set(GameCommon.iconBmpHeight)
    local button
    if disabledButton then
        button = Button(bg, disabledIcon, disabledIcon, disabledIcon, disabledIcon, 'UI_Menu_MouseDown_Sml')
        button:SetUpAltButtons( disabledIcon, disabledIcon, disabledIcon, disabledIcon )
    else
        button = Button(bg, upIconName, downIconName, overIconName, upIconName, 'UI_Menu_MouseDown_Sml')
        button:SetUpAltButtons( selectedIcon, selectedIcon, selectedIcon, selectedIcon )
    end
    --LayoutHelpers.FillParent(button, bg)
    button.Top:Set(bg.Top)
    button.Left:Set(function() return bg.Left() + 1 end)
    button._blueprintId = blueprintId
    if layoutVar == 'minimal' then
        button.Height:Set(48)
        button.Width:Set(48)
    end

    button.OnClick = function(self, modifiers)
        if not disabledButton then
            if modifiers.Left then
                -- Toggling selection of the entity
                button:OnAltToggle()

                -- Add or Remove the entity to the session selection
                if button.mAltToggledFlag then
                    AddToSessionExtraSelectList(userUnit)
                else
                    RemoveFromSessionExtraSelectList(userUnit)
                end
            end
        end
    end

    button.HandleEvent = function(self, event)
        if event.Type == 'MouseEnter' then
            local blueprint = __blueprints[self._blueprintId]
            import('/lua/modules/ui/game/unitviewDetail.lua').Show(blueprint)
        end
        if event.Type == 'MouseExit' then
            import('/lua/modules/ui/game/unitviewDetail.lua').Hide()
        end
        Button.HandleEvent(self, event)
    end

    return bg
end

-- not standard so use our own button create code
function CreateAssistingEntityButton(parent, blueprintId, blueprint, userUnit)
    local iconName, upIconName, downIconName, overIconName = GameCommon.GetCachedUnitIconFileNames(blueprint)
    local bg = Bitmap(parent)   --todo put land/air/sea background here
    bg:SetSolidColor('black')
    bg.Width:Set(GameCommon.iconBmpWidth)
    bg.Height:Set(GameCommon.iconBmpHeight)

    local button = Button(bg, upIconName, overIconName, downIconName, upIconName, 'UI_Menu_MouseDown_Sml')
    --LayoutHelpers.FillParent(button, bg)
    button.Top:Set(bg.Top)
    button.Left:Set(function() return bg.Left() + 1 end)
    button._blueprintId = blueprintId
    if layoutVar == 'minimal' then
        button.Height:Set(48)
        button.Width:Set(48)
    end
    local canAssist = userUnit:IsAutoMode()

    local text
    if canAssist then
        text = UIUtil.CreateText(button, "<LOC CONSTRUCT_0005>Assisting", 12, "Arial Bold")
    else
        text = UIUtil.CreateText(button, '', 12, "Arial Bold")
    end

    text:SetDropShadow(true)
    LayoutHelpers.AtHorizontalCenterIn(text, button)
    LayoutHelpers.AtTopIn(text, button)
    bg.text = text

    button.OnClick = function(self, modifiers)
        if modifiers.Left then
            SelectUnits( { userUnit } )
        end

        if modifiers.Right then
            if canAssist then
                canAssist = false
                userUnit:ProcessInfo('SetAutoMode', 'false')
                bg.text:SetText('')
            else
                canAssist = true
                userUnit:ProcessInfo('SetAutoMode', 'true')
                bg.text:SetText('Assisting')
            end
        end
    end

    return bg
end

-- EntityCategoryFilterDown(categories.TECH1, buildableCategories)

function EnhancementPopulate(slot)
    local filteredEnhancements = {}
    for i, v in filteredTechTable[5] do
        if v['Slot'] == slot then
            filteredEnhancements[i] = v
        end
    end
    techLevelSelected = filteredEnhancements
    buildingEnhancements = true
    PopulateGridButtons(slot)
end

function ToggleEnhancementWindow(state, enhancementWindow)
    if state then
        if UIUtil.currentLayout == 'bottom' then
            controls.unitsBGCapMin.Left:Set(function() return enhancementWindow.Right() - 23 end)
        end
    else
        if tempBuildableCategories and currentSelection and currentEntityList and UIUtil.currentLayout != 'minimal' then
            SetAvailableGridButtons(tempBuildableCategories, currentSelection, currentEntityList)
        end
        if UIUtil.currentLayout == 'bottom' then
            LayoutHelpers.AtLeftIn(controls.unitsBGCapMin, controls.constructionGroup, -2)
            LayoutHelpers.ResetRight(controls.unitsBGCapMin)
        end
    end
end

function PopulateGridButtons(inSlot)
    controls.unitGrid:DeleteAndDestroyAll(true)
    ClearSessionExtraSelectList()
    if GameMain.gameUIHidden then return end
    
    if not buildingEnhancements and enhancementBool then
        controls.enhanceButton:Enable()
    else
        controls.enhanceButton:Disable()
    end

    local totalCount = 0
    currentSelection = ValidateUnitsList(currentSelection)    
    if table.getn(currentSelection) >= 1 then
        -- Check for setting up the transport content list
        if tempBuildableCategories then
            if EntityCategoryEmpty(tempBuildableCategories) and not buildingEnhancements then
                local userUnitList = GetAttachedUnitsList(currentSelection)
                totalCount = table.getn(userUnitList)
                if totalCount > 0 then
                    local visCols, visRows = controls.unitGrid:GetVisible()
                    controls.unitGrid:AppendCols(visCols, true)
                    controls.unitGrid:AppendRows(math.ceil(table.getn(userUnitList) / visCols), true)
        
                    for i,v in userUnitList do
                            local blueprintID = v:GetBlueprint().BlueprintId
                            local blueprint = __blueprints[blueprintID]
                            local button = CreateAttachedEntityButton(controls.unitGrid, blueprintID, blueprint, v)
                            local col = math.mod(i - 1, visCols) + 1 -- note the sub/add due to one based arrays
                            local row = math.ceil(i / visCols)
                            controls.unitGrid:SetItem(button, col, row, true)
                    end
                end
            -- Check for setting up the buildable categories list
            else
                local assistList = GetAssistingUnitsList(currentSelection)
                local entityCount = GetTableSize(techLevelSelected)
                totalCount = entityCount + table.getn(assistList)
                local visCols, visRows = controls.unitGrid:GetVisible()
                controls.unitGrid:AppendCols(visCols, true)
                controls.unitGrid:AppendRows(math.ceil(totalCount / visCols), true)
        
                if buildingEnhancements then
                    local tempEntityID = currentSelection[1]:GetEntityId()
                    local existingEnhancements = import('/lua/modules/enhancementcommon.lua').GetEnhancements(tempEntityID)
                    local index = 1
                    local restrictedList = import('/lua/modules/enhancementcommon.lua').GetRestricted()
                    for i, v in techLevelSelected do
                        local isRestricted = false
                        if restrictedList then
                            for h, x in restrictedList do
                                if x == i then
                                    isRestricted = true
                                    break
                                end
                            end
                        end
                        if not isRestricted then
                            if existingEnhancements == nil then
                                if v['Prerequisite'] == nil then
                                    local button = CreateEnhancementButton(controls.unitGrid, i, v)
                                    local col = math.mod(index - 1, visCols) + 1 -- note the sub/add due to one based arrays
                                    local row = math.ceil(index / visCols)
                                    controls.unitGrid:SetItem(button, col, row, true)
                                    index = index + 1
                                end
                            else
                                if existingEnhancements[inSlot] != nil then
                                    if existingEnhancements[inSlot] == v["Prerequisite"] then
                                        local button = CreateEnhancementButton(controls.unitGrid, i, v)
                                        local col = math.mod(index - 1, visCols) + 1 -- note the sub/add due to one based arrays
                                        local row = math.ceil(index / visCols)
                                        controls.unitGrid:SetItem(button, col, row, true)
                                        index = index + 1
                                    end
                                else
                                    if v['Prerequisite'] == nil then
                                        local button = CreateEnhancementButton(controls.unitGrid, i, v)
                                        local col = math.mod(index - 1, visCols) + 1 -- note the sub/add due to one based arrays
                                        local row = math.ceil(index / visCols)
                                        controls.unitGrid:SetItem(button, col, row, true)
                                        index = index + 1
                                    end
                                end
                            end
                        end
                    end
                else
                    if totalCount> 0 then
                        for i,v in techLevelSelected do
                            local blueprint = __blueprints[v]
                            local button = CreateBuildableButton(controls.unitGrid, v, blueprint, currentSelection)
                            local col = math.mod(i - 1, visCols) + 1 -- note the sub/add due to one based arrays
                            local row = math.ceil(i / visCols)
                            controls.unitGrid:SetItem(button, col, row, true)
                        end
                        for i,v in assistList do
                            local index = entityCount + i
                            local blueprintID = v:GetBlueprint().BlueprintId
                            local blueprint = __blueprints[blueprintID]
                            local button = CreateAssistingEntityButton(controls.unitGrid, blueprintID, blueprint, v)
                            local col = math.mod(index - 1, visCols) + 1 -- note the sub/add due to one based arrays
                            local row = math.ceil(index / visCols)
                            controls.unitGrid:SetItem(button, col, row, true)
                        end
                    end
                end
            end
            controls.unitGrid:EndBatch()
        
            -- Determine the check status of the "infinite loop" button
            local onlyFactories = true
            currentInfiniteQueueCheckStatus = false
        
            for i,v in currentSelection do
                if v:IsRepeatQueue() then
                    currentInfiniteQueueCheckStatus = true
                end
                if not v:IsInCategory('FACTORY') then
                    onlyFactories = false
                end
            end
        
            if onlyFactories then
                Queue.SetInfiniteQueueToggle(true)
                Queue.SetInfinityCheck(currentInfiniteQueueCheckStatus)
            else
                Queue.SetInfinityCheck(false)
            end
        
            local visCols, visRows = controls.unitGrid:GetVisible()
            if (visCols * visRows) >= totalCount then
                controls.scrollUnitsBackButton:Disable()
                controls.scrollUnitsFwdButton:Disable()
            else
                controls.scrollUnitsBackButton:Enable()
                controls.scrollUnitsFwdButton:Enable()
            end
        end
    end
end


function SetAvailableGridButtons(buildableCategories, unitList, entityList)
    buildingEnhancements = false
    currentEntityList = table.deepcopy(entityList)
    tempBuildableCategories = buildableCategories
    currentSelection = unitList
    local buttonToSet = false
    local buildingUnitBP = false

    SetConstructionState(false)

    -- remove upgrades from entity list
    collapseArray(entityList, function(entity) return (__blueprints[entity].General.Category == 'Upgrade') end)
    filteredTech1Units = false
    filteredTech2Units = false
    filteredTech3Units = false
    filteredTech4Units = false
    filteredFactories = {}
    filteredEnhancements = false

    filteredTechTable[1] = false
    filteredTechTable[2] = false
    filteredTechTable[3] = false
    filteredTechTable[4] = false
    filteredTechTable[5] = false

    local index = 1
    for i, v in entityList do
        local blueprint = __blueprints[v]
        for h, x in blueprint.Categories do
            if x == 'CONSTRUCTIONSORTDOWN' then
                filteredFactories[index] = v
                table.remove(entityList, i)
            end
        end
    end

    filteredTech1Units = EntityCategoryFilterDown(categories.TECH1, entityList)
    filteredTech2Units = EntityCategoryFilterDown(categories.TECH2, entityList)
    filteredTech3Units = EntityCategoryFilterDown(categories.TECH3, entityList)
    filteredTech4Units = EntityCategoryFilterDown(categories.EXPERIMENTAL, entityList)

    for i, v in filteredFactories do
        local blueprint = __blueprints[v]
        for h, x in blueprint.Categories do
            if x == 'TECH2' then
                table.insert(filteredTech1Units, v)
            elseif x == 'TECH3' then
                table.insert(filteredTech2Units, v)
            end
        end
    end
    
    if table.getn(currentSelection) == 1 then
        if currentSelection[1]:IsDead() then return end
        buildingUnitBP = currentSelection[1]:GetBlueprint()

        if buildingUnitBP["Enhancements"] then
            filteredEnhancements = buildingUnitBP["Enhancements"]
        end
    end
    
    local function SortFunc(unit1, unit2)
        local bp1 = __blueprints[unit1].BuildIconSortPriority or __blueprints[unit1].StrategicIconSortPriority
        local bp2 = __blueprints[unit2].BuildIconSortPriority or __blueprints[unit2].StrategicIconSortPriority
        if bp1 >= bp2 then
            return false
        else
            return true
        end
    end
    
    if filteredTech1Units then
        table.sort(filteredTech1Units, SortFunc)
    end
    if filteredTech2Units then
        table.sort(filteredTech2Units, SortFunc)
    end
    if filteredTech3Units then
        table.sort(filteredTech3Units, SortFunc)
    end
    if filteredTech4Units then
        table.sort(filteredTech4Units, SortFunc)
    end
    if filteredEnhancements then
        table.sort(filteredEnhancements, SortFunc)
    end
    
    filteredTechTable[1] = filteredTech1Units
    filteredTechTable[2] = filteredTech2Units
    filteredTechTable[3] = filteredTech3Units
    filteredTechTable[4] = filteredTech4Units
    filteredTechTable[5] = filteredEnhancements
    
    if filteredEnhancements then
        --import('/lua/modules/ui/game/orders.lua').ToggleEnhancementButton(filteredEnhancements, currentSelection[1])
        enhancementBool = true
    else
        --import('/lua/modules/ui/game/orders.lua').ToggleEnhancementButton(false)
        enhancementBool = false
    end

    if table.getn(filteredTech1Units) > 0 then
        controls.tech1Tab.checkbox:Enable()
        techLevelSelected = filteredTechTable[1]
        buttonToSet = 1
        buildingEnhancements = false
    else
        controls.tech1Tab.checkbox:Disable()
    end
    if table.getn(filteredTech2Units) > 0 then
        controls.tech2disabledOverGroup:Hide()
        controls.tech2Tab.checkbox:Enable()
        techLevelSelected = filteredTechTable[2]
        buttonToSet = 2
        buildingEnhancements = false
    else
        controls.tech2disabledOverGroup:Show()
        controls.tech2Tab.checkbox:Disable()
    end
    if table.getn(filteredTech3Units) > 0 then
        controls.tech3disabledOverGroup:Hide()
        controls.tech3Tab.checkbox:Enable()
        techLevelSelected = filteredTechTable[3]
        buttonToSet = 3
        buildingEnhancements = false
    else
        controls.tech3disabledOverGroup:Show()
        controls.tech3Tab.checkbox:Disable()
    end
    if table.getn(filteredTech4Units) > 0 then
        controls.tech4disabledOverGroup:Hide()
        controls.tech4Tab.checkbox:Enable()
    else
        controls.tech4disabledOverGroup:Show()
        controls.tech4Tab.checkbox:Disable()
    end


    if buttonToSet == 1 then
        controls.tech2Tab.checkbox:SetCheck(false)
        controls.tech3Tab.checkbox:SetCheck(false)
        controls.tech4Tab.checkbox:SetCheck(false)
        controls.tech1Tab.checkbox:SetCheck(true)
        controls.techRadioGroup.current = buttonToSet
    elseif buttonToSet == 2 then
        controls.tech1Tab.checkbox:SetCheck(false)
        controls.tech3Tab.checkbox:SetCheck(false)
        controls.tech4Tab.checkbox:SetCheck(false)
        controls.tech2Tab.checkbox:SetCheck(true)
        controls.techRadioGroup.current = buttonToSet
    elseif buttonToSet == 3 then
        controls.tech2Tab.checkbox:SetCheck(false)
        controls.tech1Tab.checkbox:SetCheck(false)
        controls.tech4Tab.checkbox:SetCheck(false)
        controls.tech3Tab.checkbox:SetCheck(true)
        controls.techRadioGroup.current = buttonToSet
    end

    PopulateGridButtons()
end

function NewTech(UnitList)
    newTechUnits = UnitList
end

-- given a tech level, sets that tech level, returns false if tech level not available
function SetCurrentTechTab(techLevel)
    if techLevel == 1 and controls.tech1Tab.checkbox:IsDisabled() then
        return false
    elseif techLevel == 2 and controls.tech2Tab.checkbox:IsDisabled() then
        return false
    elseif techLevel == 3 and controls.tech3Tab.checkbox:IsDisabled() then
        return false
    elseif techLevel == 4 and controls.tech4Tab.checkbox:IsDisabled() then
        return false
    elseif techLevel > 4 or techLevel < 1 then
        return false
    end
    controls.techRadioGroup:Select(techLevel)
    return true
end

function GetCurrentTechTab()
    if controls.tech1Tab.checkbox and controls.tech1Tab.checkbox:IsChecked() then
        return 1
    elseif controls.tech2Tab.checkbox and controls.tech2Tab.checkbox:IsChecked() then
        return 2
    elseif controls.tech3Tab.checkbox and controls.tech3Tab.checkbox:IsChecked() then
        return 3
    elseif controls.tech4Tab.checkbox and controls.tech4Tab.checkbox:IsChecked() then
        return 4
    else
        return nil
    end
end